#!/usr/bin/python3

from distutils.core import setup

setup(
    name='refery',
    scripts=['bin/refery'],
    version='1.0',
    description='Generic functional testing',
    author='Rostan Tabet',
    author_email='rostan.tabet@gmail.com',
)
